with text_io; use text_io;

package body pimage is

-- instanciation of integer_io package for integer
package p_integer is new integer_io(integer);
use p_integer;

--------------------------------------------------------
---- main_function: take integer as parameter and return 
----                an integer
--------------------------------------------------------

function main_function( a:in integer) return integer is

begin

return 2*a;

end main_function;

-----------------------------------------------------------
---- procedure write_int: write integer passed as parameter
-----------------------------------------------------------

procedure write_int( a: in integer ) is

begin

put(a,0);

end write_int;

-----------------------------------------------------------------------
---- function get_blue: return value of primary blue color from a pixel
---- parameter: p of pixel type
---- return c of color_prim type
-----------------------------------------------------------------------

function get_blue( p: in pixel ) return color_prim is
begin
return color_prim(p);
end get_blue;

-------------------------------------------------------------------------
---- function get_green: return value of primary green color from a pixel
---- parameter: p of pixel type
---- return c of color_prim type
-------------------------------------------------------------------------

function get_green( p: in pixel ) return color_prim is
begin
return color_prim(p/2**8);
end get_green;

---------------------------------------------------------------------
---- function get_red: return value of primary red color from a pixel
---- parameter: p of pixel type
---- return c of color_prim type
---------------------------------------------------------------------

function get_red( p: in pixel ) return color_prim is
begin
return color_prim(p/2**16);
end get_red;

--------------------------------------------------------
---- function which return a color from 3 primary colors
---- parameters: blue, green, red the 3 primary colors
---- return the color coded on 32 bits
--------------------------------------------------------

function prim_to_color( blue,green,red: in color_prim ) return pixel is

col: pixel;

begin
col:=pixel(blue+green*2**8+red*2**16);
return col;
end prim_to_color;

------------------------------------------------------
---- procedure create_im_test: create a test image for 
----                           compression
---- parameters: created image and n its dimension
------------------------------------------------------

procedure create_im_test( im: out image; n: in integer ) is

begin

for i in 1..n loop
    for j in 1..n loop
      im(i,j):=255;
    end loop;
end loop;

for i in 1..n/2 loop
    for j in 1..n/2 loop
          im(i,j):=0;    
    end loop;
end loop;

for i in n/4+1..n/2 loop
      for j in n/4+1..n/2 loop
            im(i,j):=100;    
      end loop;
end loop;

for i in 1..n/4 loop
      for j in 3*n/4..n loop
            im(i,j):=30;    
      end loop;
end loop;

for i in 1..n/8 loop
      for j in 1..n/8 loop
            im(i,j):=10;    
      end loop;
end loop;

for i in 7*n/8..n loop
      for j in 7*n/8..n loop
            im(i,j):=20;    
      end loop;
end loop;

end create_im_test;

----------------------------------------------------
---- procedure load_image: load an image from a file
---- parameters: im image to load, n its dimension
----------------------------------------------------

procedure load_image( im: out image; n: out pixel ) is

filename: string(1..4);
f: pack_pixel.file_type;
dim:pixel;

begin

put("Enter a filename(4 characters): ");
get(filename);
skip_line;
open(f,in_file,filename);
read(f,dim);
n:=dim;

for i in 1..n loop
      for j in 1..n loop
        read(f,im(integer(i),integer(j)));
      end loop;
end loop;
close(f);

end load_image;

--------------------------------------------------
---- procedure save_image: save an image into file
---- parameters: im image to save; n its dimension
--------------------------------------------------

procedure save_image ( im: in image; n: in pixel ) is

filename: string(1..4);
f: pack_pixel.file_type;

begin

put("Enter a filename(4 characters): ");
get(filename);
skip_line;
create(f,out_file,filename);
write(f,n);
for i in 1..n loop
      for j in 1..n loop
            write(f,im(integer(i),integer(j)));
      end loop;
end loop;
close(f);      
      
end save_image;

------------------------------------------------------------------
---- procedure save_imagec: save a compressed image into file from
----                        a tree
---- parameters: a built tree, n the dimension of image
------------------------------------------------------------------

procedure save_imagec( a: in tree_im ; n: in pixel ) is

filename: string(1..4);
f: pack_pixel.file_type;

begin

put("Enter a filename(4 characters): ");
get(filename);
skip_line;
create(f,out_file,filename);
write(f,n);
tree_to_imc(a,f);
close(f);      

end save_imagec;

------------------------------------------------------------------
---- procedure display_image: display an image doing an explicit
----                          conversion of decimal value of pixel
----                          into character
---- parameters: im image to display, n its dimension
------------------------------------------------------------------

procedure display_image( im: in image; n: in integer ) is

begin

new_line;
put("Dimension of image is");
put(n);
new_line;
new_line;
for i in 1..n loop
    for j in 1..n loop
        put(integer(im(integer(i),integer(j))));
    end loop;
    new_line;
end loop;
new_line;
    
end display_image;

----------------------------------------------------------------
---- procedure load_im_encod : build a tree from compressed 
----                           sequence
---- parameters: a the rebuilt tree, n the dimension of original 
----             image
----------------------------------------------------------------

procedure load_im_encod( a: out tree_im ; n: out pixel ) is

filename: string(1..4);
f: pack_pixel.file_type;
-- dimension of original image
dim:pixel;

begin

put("Enter a filename(4 characters): ");
get(filename);
skip_line;
open(f,in_file,filename);
-- we read dimension
read(f,dim);
n:=dim;
a:=imc_to_tree(f);
close(f);

end load_im_encod;

-------------------------------------------------------------------
---- function col_aver: compute average color of image from average
----                    of primary colors
---- parameters: im of image type
---- return a color
-------------------------------------------------------------------

function col_aver( im: in image; n: in integer ) return pixel is

blue_aver: integer;
green_aver: integer;
red_aver: integer;

begin

blue_aver:=0;
green_aver:=0;
red_aver:=0;

for i in 1..n loop
    for j in 1..n loop
        blue_aver:=blue_aver+get_blue(im(i,j));
	green_aver:=green_aver+get_green(im(i,j));
        red_aver:=red_aver+get_red(im(i,j));
    end loop;
end loop;    

blue_aver:=blue_aver/n**2;
green_aver:=green_aver/n**2;
red_aver:=red_aver/n**2;

return prim_to_color(blue_aver,green_aver,red_aver);

end col_aver;

---------------------------------------------------------------------
---- function is_homogeneous: function which checks if area of size n 
----                          homogeneous. return a boolean
---- parameters: im image
----             size n of a area
----             tol tolerance for homogeneity
---------------------------------------------------------------------

function is_homogeneous( im: in image; n: in integer; tol: in intensity ) return boolean is

result: boolean;
val,i,j: integer;
c_aver: integer;

begin
result:=true;

case tol is
     -- maximal tolerance allowed, image is considered homegeneous
     when intensity'last =>  return true;
     -- minimal tolerance, we compare first point with all others
     when intensity'first => val:=integer(im(1,1)); 
                             i:=1;
                             while i<=n and result loop
			           j:=1;
			           while j<=n and result loop
				         result:=(im(i,j)=pixel(val));
					 j:=j+1;
				   end loop;
		             i:=i+1;
			     end loop;
			     return result;
      -- general case: we compute the average color and we extract the
      -- 3 average primary colors that we compare to image pixels with
      -- condition : abs(average primary color) <= tolerance
      when others => c_aver:=integer(col_aver(im,n));
                     i:=1;
		     while i<=n and result loop
		           j:=1;
			   while j<=n and result loop
result:=(abs(get_blue(im(i,j))-get_blue(pixel(c_aver)))<=tol) and
(abs(get_green(im(i,j))-get_green(pixel(c_aver)))<=tol) and (abs(get_red(im(i,j))-get_red(pixel(c_aver)))<=tol); 			     
				 j:=j+1;
			   end loop;
			   i:=i+1;
		     end loop;      
                     return result;
end case;
       	               
end is_homogeneous;

-------------------------------------------------------------
---- function imc_to_tree: return a tree from of a compressed
----                       image
---- parameters: f the logic name of compressed file
-------------------------------------------------------------

function imc_to_tree( f: in pack_pixel.file_type ) return tree_im is

-- children of quaternary tree
a,a_nw,a_ne,a_sw,a_se: tree_im;
v: pixel;

begin

read(f,v);
-- if value equals -1 (coded by 2**24), we create a leaf
if v /= 2**24 then
   a:=Qt_Create_Leaf(integer(v));
  else
   a_nw:=imc_to_tree(f);
   a_ne:=imc_to_tree(f);
   a_sw:=imc_to_tree(f);
   a_se:=imc_to_tree(f);
   Qt_Build(-1,a_nw,a_ne,a_sw,a_se,a);
end if;
return a;	
	  
end imc_to_tree;

----------------------------------------------------------------
----- procedure tree_to_imc: allow to convert tree to compressed
-----                        sequence
----- parameters: a the built tree, f the logic name of file
----------------------------------------------------------------

procedure tree_to_imc( a: in tree_im; f: in pack_pixel.file_type ) is

begin

if Qt_Empty(a) then
   null;
  else
   if Qt_Value(a)=-1 then
      write(f,2**24);
     else
      write(f,pixel(Qt_Value(a)));
   end if;   
   tree_to_imc(Qt_Child(a,1),f);
   tree_to_imc(Qt_Brother(a,1),f);
end if;   

end tree_to_imc;


-------------------------------------------------------------------
---- procedure tree_to_im: allow to reconstruct original image from
----                       compressed image
---- parameters: a the rebuilt tree, im image to build
----             i_b,i_f,j_b,j_f the indexed (begin and final)
----             n the dimension of original image
-------------------------------------------------------------------

procedure tree_to_im ( a: in tree_im ; im: out image ; i_b,i_f,j_b,j_f: in integer ) is

begin

-- terminal case
if Qt_Empty(a) then
   null;-- general case
  else if Qt_Value(a) /= -1 then
          for i in i_b..i_f loop
	      for j in j_b..j_f loop
	          im(i,j):=pixel(Qt_Value(a));
		end loop;
	  end loop;       
         else -- we browse the children
	  tree_to_im(Qt_North_West(a),im,i_b,i_f-(i_f-i_b+1)/2,j_b,j_f-(j_f-j_b+1)/2);
	  tree_to_im(Qt_North_East(a),im,i_b,i_f-(i_f-i_b+1)/2,j_f-(j_f-j_b+1)/2+1,j_f);
	  tree_to_im(Qt_South_West(a),im,i_f-(i_f-i_b+1)/2+1,i_f,j_b,j_f-(j_f-j_b+1)/2);
	  tree_to_im(Qt_South_East(a),im,i_f-(i_f-i_b+1)/2+1,i_f,j_f-(j_f-j_b+1)/2+1,j_f);
        end if;
end if;
	
end tree_to_im;	 	  
          
------------------------------------------------------------------
---- procedure im_to_tree: crate a tree from an image 
---- parameters: im the image, n its dimesnion, a the created tree
----             tol the tolerance used by function is_homogeneous
------------------------------------------------------------------	  

procedure im_to_tree( im: in image; n: in integer; a: out tree_im; tol: in intensity) is

im_nw,im_ne,im_sw,im_se: image(1..n/2,1..n/2);
a_nw,a_ne,a_sw,a_se: tree_im;

begin
-- case where dimension equals 1
if n=1 then
   a:=Qt_Create_Leaf(integer(im(1,1)));
   -- if it is homogeneous, we create a leaf with average color
  else if is_homogeneous(im,n,tol) then
          a:=Qt_Create_Leaf(integer(col_aver(im,n)));
         else -- we build the tree dividing dimensions by two
           for i in 1..n/2 loop
	       for j in 1..n/2 loop
	           im_nw(i,j):=im(i,j);
		   im_ne(i,j):=im(i,j+n/2);
		   im_sw(i,j):=im(i+n/2,j);
		   im_se(i,j):=im(i+n/2,j+n/2);
	       end loop;
	   end loop;    	      
          
           im_to_tree(im_nw,n/2,a_nw,tol);
	   im_to_tree(im_ne,n/2,a_ne,tol);
	   im_to_tree(im_sw,n/2,a_sw,tol);
	   im_to_tree(im_se,n/2,a_se,tol);
	   Qt_Build(-1,a_nw,a_ne,a_sw,a_se,a);	    
        
       end if;
end if;       
	
end im_to_tree;

----------------------------------------------------------
----- procedure thresh_im: perform a threshold on image
----- parameters: im the image to process, n its dimension
-----             threshold: the threshold
----------------------------------------------------------

procedure thresh_im( im: in out image ; n: in integer ; threshold: in integer ) is

begin

for i in 1..n loop
    for j in 1..n loop
        if (integer(im(i,j))<= threshold) then
	   im(i,j):=0;
	  else
	    im(i,j):=1;
	end if;    
	end loop;	
end loop;

end thresh_im;

end pimage;

