with sequential_io;
with tree_quater;

generic 

with procedure write( a: in integer );
with function function_gen( a: in integer ) return integer;

package pimage is

-- instanciation of tree_quater package
package p_tree_q is new tree_quater(integer,write,function_gen);
use p_tree_q;

-- subtype tree_im 
subtype tree_im is tree_quat;
-- definition of primary color
subtype color_prim is integer range 0..255;
-- definition of pixel: value coded on 32 bits
type pixel is mod 2**32;

package pack_pixel is new sequential_io(pixel); 
use pack_pixel;

-- image array
type image is array (integer range <>,integer range <>) of pixel;
type image_comp is array (1..2**18) of pixel;
subtype intensity is integer range 0..100;
-- function for getting primary colors from a color
function get_blue( p: in pixel ) return color_prim;
function get_green( p: in pixel ) return color_prim;
function get_red( p: in pixel ) return color_prim;

-- function to compute average color of an image
function col_aver( im: in image; n: in integer ) return pixel;
-- function which returns a pixel from 3 primary colors
function prim_to_color( blue,green,red: in color_prim ) return pixel;

-- procedure for creation of a compressed image and its decompression
-- and input/output
procedure load_image( im: out image; n: out pixel );
procedure save_image( im: in image; n: in pixel );
procedure display_image ( im: in image; n: integer );
procedure display_tree( a: in tree_im ) renames Qt_Display;
procedure load_im_encod( a: out tree_im ; n: out pixel );
function imc_to_tree( f: in pack_pixel.file_type ) return tree_im;
procedure create_im_test( im: out image; n: in integer );
procedure im_to_tree( im: in image; n: in integer; a: out tree_im; tol: in intensity );
procedure tree_to_imc( a: in tree_im; f: in pack_pixel.file_type );
procedure save_imagec( a: in tree_im ; n: in pixel );
procedure tree_to_im ( a: in tree_im ; im: out image ; i_b,i_f,j_b,j_f: in integer );
procedure thresh_im( im: in out image ; n: in integer ; threshold: in integer );
end pimage;
