with text_io;use text_io;

package body tree_nary is

------------
-- Creation
------------

------------------------------------------------------------------
---- function Nt_Create_Empty: create and return an empty tree
---- parameters: none
------------------------------------------------------------------

function Nt_Create_Empty return tree_nr is

begin

return null;

end Nt_Create_Empty;

------------------------------------------------------------------
---- function Nt_Create_Leaf: create and return a n-ary tree with 
---- a value, without child, no brother, no father
---- parameters: value : value to put in tree
------------------------------------------------------------------

function Nt_Create_Leaf ( value : in T ) return tree_nr is

a: tree_nr;

begin

a:=new node;
a.val:=value;
a.first_child:=null;
a.brother:=null;
a.father:=null;
return a;

end Nt_Create_Leaf;

---------------
-- Consultation
---------------

--------------------------------------------------------------------
---- fucntion Nt_Empty: check if a tree is empty or not, return a 
---- boolean
---- parameters: n-ary tree a
--------------------------------------------------------------------         

function Nt_Empty ( a: in tree_nr ) return boolean is

begin

return (a=null);

end Nt_Empty;

-------------------------------------------------------------------------
---- function Nt_Value: return the value of the n-ary tree root
---- parameters: n-ary tree a
---- post-conditions: if tree is null, we raise tree_empty exception
-------------------------------------------------------------------------

function Nt_Value ( a: in tree_nr ) return T is

begin

return (a.val);
exception
when constraint_error => raise tree_empty;

end Nt_Value;

---------------------------------------------------------------------------------
---- function Nt_Father: return father tree of a tree
---- parameters: n-ary tree a
---- post-conditions: if tree is empty, we raise tree_empty exception
----                  if father does not exist, we raise relation_empty exception 
---------------------------------------------------------------------------------

function Nt_Father ( a: in tree_nr ) return tree_nr is

begin

if Nt_Empty(a) then
   raise tree_empty;
  else if a.father=null then 
          raise relation_empty;
         else
          return a.father;
       end if;
end if;

end Nt_Father;

-------------------------------------------------------------------------------------
---- function Nt_Child: return n-th child of a tree
---- parameters: n-ary tree a
----             n which represents n-th child
---- post-conditions: if tree is empty, we raise tree_empty exception
----                  if n-th child does not exist, we raise relation_empty exception  
-------------------------------------------------------------------------------------

function Nt_Child ( a: in tree_nr; n: in integer ) return tree_nr is

tree_curr: tree_nr;

begin

if Nt_Empty(a) then
   raise tree_empty;
  else 
     tree_curr:=a.first_child;     
     for i in 1..n-1 loop
       tree_curr:=tree_curr.brother;
     end loop;
     return tree_curr; 
end if;

exception
     when constraint_error => raise relation_empty;
end Nt_Child;

---------------------------------------------------------------------------------------
---- function Nt_Brother: return n-th brother of a tree
---- parameters: n-ary tree a
----             n which represents n-th brother
---- post-conditions: if tree is empty, we raise tree_empty exception
----                  if n-th brother does not exist, we raise relation_empty exception  
---------------------------------------------------------------------------------------

function Nt_Brother ( a: in tree_nr; n: in integer ) return tree_nr is

tree_curr: tree_nr;

begin

if Nt_Empty(a) then
   raise tree_empty;
  else 
     tree_curr:=a;
     for i in 1..n loop
       tree_curr:=tree_curr.brother;
     end loop;
     return tree_curr; 
end if;

exception
     when constraint_error => raise relation_empty;
end Nt_Brother;

---------------------------------------------------------------------
---- procedure Nt_Display: display all tree content
---- parameters: n-ary tree a
---- post-conditions: if tree is empty, we raise tree_empty exception
---------------------------------------------------------------------

procedure Nt_Display( a: in tree_nr ) is

begin
	
if Nt_Empty(a) then
   raise tree_empty;
else
	 put("    ");
   Nt_Display_Aux(a,"    ");
end if;

end Nt_Display;

---------------------------------------------------------------------
---- procedure Nt_Display_aux: recursive procedure which displays all
----                           tree children and brothers values
---- parameters: n-ary tree a
---------------------------------------------------------------------

procedure Nt_Display_Aux( a: in tree_nr ; str_shift: in string ) is

str_inc: string(1..4);

begin

if a=null then
   null;
  else
   write(a.val); 
   new_line;
   
   if (not Nt_Empty(a.first_child)) then
      put(str_shift&"|"); 
      new_line;
      put(str_shift&"+-- "); 
      if (not Nt_Empty(a.brother)) then
         str_inc:="|   ";
	else
	 str_inc:="    ";
      end if;
   
      -- we display child
      Nt_Display_Aux(a.first_child,str_shift&str_inc);
   end if;   
   
   if (not Nt_Empty(a.brother)) then
      put(str_shift);
      -- we display brothers
      Nt_Display_Aux(a.brother,str_shift);      
   end if;
   
end if;   

end Nt_Display_Aux;

-----------------------------------------------------------------------
---- function Nt_Search: search for a value in a tree
----                     return the tree whose value is located at root
----                     node if it is found
----                     return null if value is not found
----------------------------------------------------------------------

function Nt_Search ( a: in tree_nr; e: in T ) return tree_nr is 

-- current tree
tree_curr :tree_nr;
-- boolean to keep on or not searching
stop: boolean; 
begin

if Nt_Empty(a) then
   raise tree_empty;
  -- terminal case
  else if a.val=e then
	  return a;
	 -- general case: we use Nt_Search_Aux auxiliary procedure
	 -- tree_curr is returned 
	 else 
          stop:=false;
	  tree_curr:=Nt_Create_Empty;
	  Nt_Search_Aux(a.first_child,e,tree_curr,stop);
	  return tree_curr;
       end if;
end if;

end Nt_Search;

---------------------------------------------------------------------------------
---- procedure Nt_Search_Aux : recursive procedure which searches a value firstly
----                           on child relation and then brothers relation
---- parameters: n-ary tree tree, value to search e, current tree tree_current to
----             return, boolean stop which can stop searching
---- post-conditions: return null if value is not found, otherwise return tree
---------------------------------------------------------------------------------

procedure Nt_Search_Aux ( tree :in tree_nr ; e: in T ; tree_curr: out tree_nr ; stop: in out boolean ) is
     
begin

-- terminal case
if tree=null then
   null;
  else if tree.val=e then
          stop:=true;
          tree_curr:=tree;
	 -- we search firstly on child relation
         else if not stop then 
                 Nt_Search_Aux(tree.first_child,e,tree_curr,stop);
                else null;
              end if;
              -- then we search on brothers relation
              if not stop then
                 Nt_Search_Aux (tree.brother,e,tree_curr,stop);
                else null;
              end if;
       end if;     
end if;    

end Nt_Search_Aux;

---------------------------------------------------------------
---- function Nt_Number_Children: return the number of children
----                              at first level of a tree
---- parameters: n-ary tree a
---- post-conditions: if tree is null, return 0
---------------------------------------------------------------

function Nt_Number_Children( a: in tree_nr ) return integer is

n:integer;
tree_curr:tree_nr;

begin

if nt_Empty(a) then
   raise tree_empty;
 else if a.first_child /= null then
         n:=1; 
         tree_curr:=a.first_child; 
         while tree_curr.brother /= null loop
               n:=n+1;
               tree_curr:=tree_curr.brother;
         end loop;
         return n;
        else
	 return 1; 
       end if;
end if;
       
end Nt_Number_Children;

----------------------------------------------------------
---- function Nt_Is_Leaf: check if a tree is a leaf node
---- parameters: n-ary tree a
---- post-conditions: return true if tree is a leaf node,
----                  otherwise false
----------------------------------------------------------

function Nt_Is_Leaf( a: in tree_nr )return boolean is
 
begin

if a.first_child=null then
   return true;
  else return false;
end if;

exception
	when constraint_error => raise tree_empty;
	
end Nt_Is_Leaf;
     
----------------------------------------------------------
---- function Nt_Is_Root: check if a tree has no father
---- parameters: n-ary tree a
---- post-conditions: return true if tree has no father,
----                  otherwise false
----------------------------------------------------------

function Nt_Is_Root( a: in tree_nr )return boolean is
 
begin

if a.father=null then
   return true;
  else return false;
end if;

exception
	when constraint_error => raise tree_empty;
	
end Nt_Is_Root;
     
----------------     
-- Modification   
----------------  

-----------------------------------------------------------
---- procedure Nt_Change_Value: change value at the root of 
----                            a tree
---- parameters: n-ary tree a and value a of T type
---- post-conditions: if tree is empty, we raise tree_empty
----                  exception
-----------------------------------------------------------

procedure Nt_Change_Value( a:in out tree_nr; e: in T) is

begin

a.val:=e;
exception
    when constraint_error => raise tree_empty;
    
end Nt_Change_Value;

--------------------------------------------------------------
---- procedure Nt_Insert_Child: insert a tree without brothers
----                            at first child position of a 
----                            tree a. Old child of a becomes
----                            the first brother of inserted
----                            tree
---- parameters: n-ary tree tree 
----             n-ary tree tree_i which is the tree to insert
---- post-conditions: if tree is empty, we raise tree_empty
----                  exception
--------------------------------------------------------------

procedure Nt_Insert_Child ( tree: in out tree_nr ; tree_i: in out tree_nr ) is
    
begin

if Nt_Empty(tree) then
   raise tree_empty;
-- if tree to insert is empty, we do noting
  else if tree_i=null then
          null;
	 -- 2 possible cases: tree has a child or not 
         else if tree.first_child=null then
        	 tree.first_child:=tree_i;
	 -- we insert tree_i at first child position and old child
	 -- become the first brother 	 
        	else 
                 tree_i.brother:=tree.first_child;
                 tree_i.father:=tree;
                 tree.first_child:=tree_i;
              end if;
       end if;
end if;
    
end Nt_Insert_Child;

----------------------------------------------------------------
---- procedure Nt_Insert_Brother: insert a tree without brothers
----                              at first brother position of a 
----                              tree a
---- parameters: n-ary tree
----             n-ary tree tree_i which is the tree to insert
---- post-conditions: if tree is empty, we raise tree_empty
----                  exception
----------------------------------------------------------------

procedure Nt_Insert_Brother( tree: in out tree_nr ; tree_i: in out tree_nr ) is
    
begin

if Nt_Empty(tree) then
   raise tree_empty;
  else if tree_i=null then
          null;
         else              
          tree_i.father:=tree.father;
	  tree_i.brother:=tree.brother;
          tree.brother:=tree_i;
       end if;
end if;
   
end Nt_Insert_Brother;

------------------------------------------------------------------
---- procedure Nt_Delete_Child: delete the n-th child of a tree a;
----                            if first child is deleted, the 
----                            first brother of deleted child 
----                            becomes the first child of a 
---- parameters: n-ary tree a
----             n the number of child to delete
------------------------------------------------------------------

procedure Nt_Delete_Child ( a: in out tree_nr ; n: in integer ) is

-- saved trees: previous and following the child 
a_prev: tree_nr;
a_next: tree_nr; 
   
begin
    
if Nt_Empty(a) then
   raise tree_empty;
   -- if the n-th child does not exist, we do nothing
  else if n>Nt_Number_Children(a) then 
          null;
	 -- if n=1, deleting the first child
         else if n=1 then 
                 a_next:=a.first_child;
        	 a_next.father:=null;
                 a.first_child:=a_next.brother;
		 -- we save pointer on n-th child
		 -- ans pointer on child's brother to delete
        	else 
                 a_prev:=Nt_Child(a,n-1);
                 a_next:=a_prev.brother.brother;
                 a_prev.brother.brother:=null;
        	 a_prev.brother.father:=null;
                 a_prev.brother:=a_next;
               end if;
       end if;
end if;

end Nt_Delete_Child;
  
------------------------------------------------------------  
---- procedure Nt_Delete_Brother: delete the n-th brother of 
----                              a tree a. New brother of a
----                              will be the brother of 
----                              deleted tree.
---- parameters: n-ary tree a
----             n the number of brother to delete
---- post-conditions: if relation_empty is raised into 
----                  Nt_Brother, then we propagate it if 
----                  a_prev or a_prev.brother is null and 
----                  constraint_error is raised
------------------------------------------------------------  
  
procedure Nt_Delete_Brother ( a :in out tree_nr ; n: in integer) is

-- saved trees: previous and following the child 
a_prev: tree_nr;
a_next: tree_nr; 

begin
if Nt_Empty(a) then
   raise tree_empty;
  -- deleting the n-th brother
  else 
  -- previous tree before brother to delete
  -- if relation_empty is raised into Nt_Brother,
  -- it propagates
  a_prev:=Nt_Brother(a,n-1);
  -- next tree after brother to delete 
  -- if a_prev or a_prev.brother is raised, 
  -- constraint_error is raised, then it propagates
  a_next:=a_prev.brother.brother; 
  -- case where a_prev or a_prev.brother is not null
  a_prev.brother.father:=null;
  a_prev.brother.brother:=null;
  a_prev.brother:=a_next;
end if;
    
exception
   
 -- if a_prev or a_prev.brother is null, we do noting
 when constraint_error => null;
                               
end Nt_Delete_Brother;

-----------------------------------------------------
---- procedure Nt_Change: browse tree and apply value
----                      for each node of tree
---- parameters: n-ary tree a
---- post-conditions: if tree is empty, we do nothing
-----------------------------------------------------

procedure Nt_Change( a: in out tree_nr ) is

begin

if Nt_Empty(a) then
   null;
  else
   a.val:=main_function(a.val);
   Nt_Change(a.first_child);
   Nt_Change(a.brother);
end if;   

end Nt_Change;

------------------------------------------------------------
---- procedure Insert_First_Child: insert a first_child into
----                               a tree a
---- parameters: n-ary tree a
----             n-ary tree a_ins to insert
------------------------------------------------------------  

procedure Insert_First_Child ( a: in out tree_nr ; a_ins: in out tree_nr) is

a_next: tree_nr;

begin

a_next:=a.first_child;
a.first_child:=a_ins;
a_ins.father:=a;
a_ins.first_child:=a_next;
a_next.father:=a_ins;

end Insert_First_Child;

end tree_nary;

