package body tree_quater is

-- Creation

----------------------------------------------------------------
---- procedure Qt_Build: create quaternary tree from a value and
----                     and 4 child trees
---- parameters: tree_res, built tree
--               val, father value
----            a_nw, a_ne, a_sw, a_se, respectively north-west,
----            north-east, south-west and south-east children
---- post-conditions: if one of child is empty, we raise 
----                  childq_empty exception
----------------------------------------------------------------

procedure Qt_Build( value: in T; t_nw, t_ne, t_sw, t_se: in out tree_quat; tree_res: out tree_quat ) is

begin

-- if one of child is empty, we raise childq_empty exception
if (Qt_Empty(t_nw) or Qt_Empty(t_ne) or Qt_Empty(t_sw) or Qt_Empty(t_se)) then
   raise childq_empty;
-- otherwise we build quaternary tree   
  else 
   tree_res:=Qt_Create_Leaf(value);
   Nt_Insert_Child(tree_res,t_se);
   Nt_Insert_Child(tree_res,t_sw);
   Nt_Insert_Child(tree_res,t_ne);
   Nt_Insert_Child(tree_res,t_nw);

end if;
    
end Qt_Build;

-- Consultation 

--------------------------------------------------------------
---- function Qt_North_West: return the north-west child of a
----                         quaternary tree
---- parameters: a quaternary tree
--------------------------------------------------------------

function Qt_North_West( a: in tree_quat ) return tree_quat is

begin

return Nt_Child(a,1);

end Qt_North_West;  

--------------------------------------------------------------
---- function Qt_North_East: return the north-east child of a
----                         quaternary tree
---- parameters: a quaternary tree
--------------------------------------------------------------

function Qt_North_East( a: in tree_quat ) return tree_quat is

begin

return Nt_Child(a,2);

end Qt_North_East;  

--------------------------------------------------------------
---- function Qt_South_West: return the south-west child of a
----                         quaternary tree
---- parameters: a quaternary tree
--------------------------------------------------------------

function Qt_South_West( a: in tree_quat ) return tree_quat is

begin

return Nt_Child(a,3);

end Qt_South_West;  

--------------------------------------------------------------
---- function Qt_South_East: return the south-east child of a
----                         quaternary tree
---- parameters: a quaternary tree
--------------------------------------------------------------

function Qt_South_East( a: in tree_quat ) return tree_quat is

begin

return Nt_Child(a,4);

end Qt_South_East;  

end tree_quater;

