-- generic parameters:
-- values T type of quaternary tree
-- write routine
-- main_function routine used in Qt_Change

with tree_nary;

generic

type T is private;
with procedure write ( a: in T );
with function main_function ( a: in T) return T;

-- specifications part of package

package tree_quater is

--instanciation of tree_nary package
package tree_nr_gen is new tree_nary(T,write,main_function);
use tree_nr_gen;

-- tree_quat is a subtype of tree_nr
subtype tree_quat is tree_nr;

-- exception

-- empty tree
treeq_empty: exception renames tree_empty;
-- no relation
relationq_empty: exception renames relation_empty;
-- one of quaternary tree children is empty
childq_empty: exception;

-- creation

function Qt_Create_Empty return tree_quat renames Nt_Create_Empty;
function Qt_Create_Leaf ( value: in T ) return tree_quat renames Nt_Create_Leaf;
procedure Qt_Build ( value: in T; t_nw, t_ne, t_sw, t_se: in out tree_quat; tree_res: out tree_quat ); 

-- consultation 

function Qt_Empty ( a: in tree_quat ) return boolean renames Nt_Empty;
function Qt_Value ( a: in tree_quat ) return T renames Nt_Value;
function Qt_Father ( a: in tree_quat ) return tree_quat renames Nt_Father;
function Qt_Child ( a: in tree_quat; n: in integer ) return tree_quat 
renames Nt_Child;
function Qt_Brother ( a: in tree_quat; n: in integer ) return tree_quat 
renames Nt_Brother;
function Qt_North_West ( a: in tree_quat ) return tree_quat;
function Qt_North_East ( a: in tree_quat ) return tree_quat;
function Qt_South_West ( a: in tree_quat ) return tree_quat;
function Qt_South_East ( a: in tree_quat ) return tree_quat;
procedure Qt_Display ( a: in tree_quat ) renames Nt_Display;
function Qt_Search ( a: in tree_quat; e: in T ) return tree_quat renames
Nt_Search;
function Qt_Is_Leaf ( a: in tree_quat ) return boolean renames Nt_Is_Leaf;
function Qt_Is_Root ( a: in tree_quat ) return boolean renames Nt_Is_Root;

-- modification

procedure Qt_Insert_Brother ( a: in out tree_quat ; a_i: in out tree_quat )
renames Nt_Insert_Brother;
procedure Qt_Insert_Child ( a: in out tree_quat ; a_i: in out tree_quat )
renames Nt_Insert_Child;
procedure Qt_Change_Value ( a: in out tree_quat; e: in T ) renames
Nt_Change_Value;
procedure Qt_Change( a: in out tree_quat ) renames Nt_Change;

end tree_quater;

