with text_io;use text_io;
package body arbre_nr is

------------
-- Cration
------------

------------------------------------------------------------------
---- fonction An_Creer_Vide: cre et retourne un arbre vide
---- parametres: aucun
------------------------------------------------------------------

function An_Creer_Vide return arb_nr is

begin

return null;

end An_Creer_Vide;


------------------------------------------------------------------
---- fonction An_Creer_Feuille: cre et retourne un arbre n-aire
---- avec une valeur, sans fils, ni frere, ni pere
---- parametres: valeur : valeur  mettre dans l'arbre
------------------------------------------------------------------

function An_Creer_Feuille ( valeur : in T ) return arb_nr is

a: arb_nr;

begin

a:=new noeud;
a.val:=valeur;
a.premier_fils:=null;
a.frere:=null;
a.pere:=null;
return a;

end An_Creer_Feuille;


---------------
-- Consultation
---------------
         
--------------------------------------------------------------------
---- fonction An_Vide: dtecte si un arbre est vide ou non, retourne
---- un boolen.
---- parametres: arbre n-aire a 
--------------------------------------------------------------------

function An_Vide ( a: in arb_nr ) return boolean is

begin

return (a=null);

end An_Vide;

-------------------------------------------------------------------------
---- fonction An_Valeur: retourne la valeur range  la racine d'un arbre
---- paramtres: arbre n-aire a
---- post-conditions: si l'arbre est null, on leve l'exception arbre_vide
-------------------------------------------------------------------------

function An_Valeur ( a: in arb_nr ) return T is

begin

return (a.val);
exception
when constraint_error => raise arbre_vide;

end An_Valeur;

-------------------------------------------------------------------------------
---- fonction An_Pere: retourne l'arbre pre d'un arbre
---- paramtres: arbre n-aire a 
---- post-conditions: si l'arbre est vide, on leve l'exception arbre_vide
----                  si le pere n'existe pas, on leve l'exception parente_vide
-------------------------------------------------------------------------------

function An_Pere ( a: in arb_nr ) return arb_nr is

begin

if An_Vide(a) then
   raise arbre_vide;
  else if a.pere=null then 
          raise parente_vide;
         else
          return a.pere;
       end if;
end if;

end An_pere;

------------------------------------------------------------------------------------
---- fonction An_Fils: retourne l'arbre nime fils d'un arbre
---- paramtres: arbre n-aire a
----             n qui reprsente le nime fils
---- post-conditions: si l'arbre est vide, on leve l'exception arbre_vide
----                  si le nime fils n'existe pas, on leve l'exception parente_vide
-------------------------------------------------------------------------------------

function An_Fils ( a: in arb_nr; n: in integer ) return arb_nr is

arb_cour: arb_nr;

begin

if An_Vide(a) then
   raise arbre_vide;
  else 
     arb_cour:=a.premier_fils;     
     for i in 1..n-1 loop
       arb_cour:=arb_cour.frere;
     end loop;
     return arb_cour; 
end if;

exception
     when constraint_error => raise parente_vide;
end An_Fils;

------------------------------------------------------------------------------------
---- fonction An_Frere: retourne l'arbre nime frere d'un arbre
---- paramtres: arbre n-aire a
----             n qui reprsente le nime frre
---- post-conditions: si l'arbre est vide, on leve l'exception arbre_vide
----                  si le nime frre n'existe pas, on leve l'exception parente_vide
-------------------------------------------------------------------------------------

function An_Frere ( a: in arb_nr; n: in integer ) return arb_nr is

arb_cour: arb_nr;

begin

if An_Vide(a) then
   raise arbre_vide;
  else 
     arb_cour:=a;
     for i in 1..n loop
       arb_cour:=arb_cour.frere;
     end loop;
     return arb_cour; 
end if;

exception
     when constraint_error => raise parente_vide;
end An_Frere;

-------------------------------------------------------------------
---- procedure An_Afficher : affiche le contenu complet de l'arbre
---- paramtres: arbre n-aire a
---- post-conditions: si l'arbre est vide, on leve l'exception arbre_vide           
-------------------------------------------------------------------

procedure An_Afficher( a: in arb_nr ) is

begin
	
if An_Vide(a) then
   raise arbre_vide;
else
	 put("    ");
   An_Afficher_Aux(a,"    ");
end if;


end An_Afficher;

---------------------------------------------------------------------
---- procedure An_Afficher_aux : procdure rcursive qui affiche les
----  valeurs de tous les fils et tous les freres.
---- parametres: arbre n-aire a
------------------------------------------------------------------

procedure An_Afficher_Aux( a: in arb_nr ; str_decal: in string ) is

str_inc: string(1..4);

begin

if a=null then
   null;
  else  
   ecrire(a.val);
   new_line;
   
   if (not An_Vide(a.premier_fils)) then
      put(str_decal&"|");
      new_line;      
      put(str_decal&"+-- ");       
      if (not An_Vide(a.frere))  then
         str_inc:="|   ";
	else
         str_inc:="    ";
      end if;
      
      -- on affiche le fils
      An_Afficher_Aux(a.premier_fils,str_decal&str_inc);                     
   end if;
         
   if (not An_Vide(a.frere)) then
      put(str_decal); 
      -- on affiche le frere
      An_Afficher_Aux(a.frere,str_decal);
   end if; 
   
end if;   

end An_Afficher_Aux;


----------------------------------------------------------------------
---- fonction An_Rechercher: recherche une valeur dans un arbre et 
---- retourne l'arbre dont la valeur est racine si elle est trouve
---- retourne null si la valeur n'est pas trouve
----------------------------------------------------------------------

function An_Rechercher ( a: in arb_nr; e: in T ) return arb_nr is 

 -- arbre courant 
arb_cour :arb_nr;
-- booleen pour continuer ou arreter la recherche
arret: boolean; 
begin

if An_Vide(a) then
   raise arbre_vide;
  -- cas terminal 
  else if a.val=e then
	  return a;
	 -- cas gnral: on utilise la procdure auxiliaire An_Rechercher_Aux 
	 -- l'arbre  retourner est arb_cour
	 else 
          arret:=false;
	  arb_cour:=An_Creer_Vide;
	  An_Rechercher_Aux(a.premier_fils,e,arb_cour,arret);
	  return arb_cour;
       end if;
end if;

end An_Rechercher;


-----------------------------------------------------------------------------
---- procedure An_Rechercher_Aux : procdure rcursive qui recherche d'abord
---- dans la ligne des fils puis des frres. 
---- paramtres: arbre n-aire arb, e la valeur  chercher, arb_cour l'arbre 
---- courant que l'on retournera, arret le boolen qui permet d'arreter la 
---- recherche
---- post-conditions: retourne null si la valeur n'est pas trouve, sinon 
---- retourne l'arbre 
-----------------------------------------------------------------------------

procedure An_Rechercher_Aux ( arb :in arb_nr ; e: in T ; arb_cour: out arb_nr ; arret: in out boolean ) is
     
begin

 -- cas terminal
if arb=null then
   null;
  else if arb.val=e then
          arret:=true;
          arb_cour:=arb;
	 -- on recherche d'abord dans la ligne des fils
         else if not arret then 
                 An_Rechercher_Aux(arb.premier_fils,e,arb_cour,arret);
                else null;
              end if;
              --on recherche ensuite dans la ligne des frres
              if not arret then
                 An_Rechercher_Aux (arb.frere,e,arb_cour,arret);
                else null;
              end if;
       end if;     
end if;    

end An_Rechercher_Aux;



------------------------------------------------------------
---- fonction An_Nombre_Fils: retourne le nombre de fils au
---- premier niveau d'un arbre
---- parametres: a arbre n-aire
---- post-conditions: si l'arbre est null, on retourne 0
------------------------------------------------------------


function An_Nombre_Fils( a: in arb_nr ) return integer is

n:integer;
arb_cour:arb_nr;

begin

if An_Vide(a) then
   raise arbre_vide;
 else if a.premier_fils /= null then
         n:=1; 
         arb_cour:=a.premier_fils; 
         while arb_cour.frere /= null loop
               n:=n+1;
               arb_cour:=arb_cour.frere;
         end loop;
         return n;
        else
	 return 1; 
       end if;
end if;
       
end An_Nombre_Fils;


----------------------------------------------------------
---- fonction An_Est_Feuille: indique si un arbre est une 
---- feuille
---- parametres: arbre n-aire a
---- post-conditions: retourne vrai si l'arbre est une 
---- feuille, sinon retourne faux
----------------------------------------------------------

function An_Est_Feuille( a: in arb_nr )return boolean is
 
begin

if a.premier_fils=null then
   return true;
  else return false;
end if;

exception
	when constraint_error => raise arbre_vide;
	
end An_Est_Feuille;
     

----------------------------------------------------------
---- fonction An_Est_Racine: indique si un arbre est sans
---- pre
---- parametres: arbre n-aire a
---- post-conditions: retourne vrai si l'arbre est sans 
---- pre, sinon retourne faux
----------------------------------------------------------

function An_Est_Racine( a: in arb_nr )return boolean is
 
begin

if a.pere=null then
   return true;
  else return false;
end if;

exception
	when constraint_error => raise arbre_vide;
	
end An_Est_Racine;
     
----------------     
-- Modification   
----------------  

-----------------------------------------------------------
---- procdure An_Changer_Valeur: change la valeur range 
---- la racine d'un arbre
---- paramtres: arbre n-aire a et valeur e de type T
---- post-conditions: si l'arbre est vide, on leve 
---- l'exception arbre_vide
-----------------------------------------------------------


procedure An_Changer_Valeur ( a:in out arb_nr; e: in T) is

begin

a.val:=e;
exception
    when constraint_error => raise arbre_vide;
    
end An_Changer_Valeur;


------------------------------------------------------------
---- procdure An_Inserer_Fils : insre un arbre sans frre 
---- en position de premier fils d'un arbre a. L'ancien fils
---- de a devient le premier frre de l'arbre insr
---- paramtres: arb qui est l'arbre a
----             arb_i qui est l'arbre  insrer  
---- post-conditions: si l'arbre est vide, on lve l'exception
---- arbre_vide
------------------------------------------------------------


procedure An_Inserer_Fils ( arb: in out arb_nr ; arb_i: in out arb_nr ) is
    
begin

if An_Vide(arb) then
   raise arbre_vide;
-- si l'arbre  insrer est vide, on ne fait rien   
  else if arb_i=null then
          null;
	 -- 2 cas possibles: l'arbre arb a un fils ou n'a pas de fils  
         else if arb.premier_fils=null then
        	 arb.premier_fils:=arb_i;
         -- on insre arb_i en position de premier fils de arb et 
	 -- l'ancien fils devient le premier frere du fils insr 
        	else 
                 arb_i.frere:=arb.premier_fils;
                 arb_i.pere:=arb;
                 arb.premier_fils:=arb_i;
              end if;
       end if;
end if;
    
end An_Inserer_Fils;

--------------------------------------------------------------
---- procdure An_Inserer_Frere : insre un arbre sans frre 
---- en position de premier frre d'un arbre a 
---- paramtres: arb qui est l'arbre a
----             arb_i l'arbre  insrer
---- post-conditions: si l'arbre est vide, on leve l'exception
---- arbre_vide
--------------------------------------------------------------

procedure An_Inserer_Frere( arb: in out arb_nr ; arb_i: in out arb_nr ) is
    
begin

if An_Vide(arb) then
   raise arbre_vide;
  else if arb_i=null then
          null;
         else              
          arb_i.pere:=arb.pere;
	  arb_i.frere:=arb.frere;
          arb.frere:=arb_i;
       end if;
end if;
   
end An_Inserer_Frere;

----------------------------------------------------------------
---- procdure An_Supprimer_Fils: supprime le n-ime fils d'un 
---- arbre a : dans le cas o l'on supprime le premier fils de a,
---- le premier frre du fils supprim devient le fils de a
---- paramtres: arbre n-aire a
----             le numro du fils  supprimer n
----------------------------------------------------------------

procedure An_Supprimer_Fils ( a: in out arb_nr ; n: in integer ) is

-- arbre sauvegards: prcdant et suivant le fils
a_s: arb_nr; 
a_pre: arb_nr;
   
begin
    
if An_Vide(a) then
   raise arbre_vide;
   -- si le n ieme fils n'existe pas, on ne fait rien
  else if n>An_nombre_fils(a) then 
          null;
	 -- si n=1, suppression du premier fils 
         else if n=1 then 
                 a_s:=a.premier_fils;
        	 a_s.pere:=null;
                 a.premier_fils:=a_s.frere;
		 -- on sauvegarde le pointeur sur le n-1 ime fil
		 -- et le pointeur sur le frere du fils  supprimer
        	else --
                 a_pre:=An_Fils(a,n-1);
                 a_s:=a_pre.frere.frere;
                 a_pre.frere.frere:=null;
        	 a_pre.frere.pere:=null;
                 a_pre.frere:=a_s;
               end if;
       end if;
end if;

end An_Supprimer_Fils;
  
------------------------------------------------------------  
---- procedure An_Supprimer_Frere : supprime le n-ime frre  
---- d'un arbre a. Le nouveau frre de a sera le frre de 
---- l'arbre supprim.
---- paramtres: arbre n-aire a
----             le numro du frre  supprimer n                 
---- post-conditions: si l'exception parente_vide est leve
---- dans An_Frere, alors on la laisse se propager
---- si a_pre ou a_pre.frere est null, alors l'exception 
---- constraint_error est leve
------------------------------------------------------------  
  
procedure An_Supprimer_Frere ( a :in out arb_nr ; n: in integer) is

-- arbres sauvegards: prcedant et suivant le frre 
a_pre:arb_nr; 
a_s:arb_nr; 
begin
if An_Vide(a) then
   raise arbre_vide;
  -- suppression du nime frre
  else 
-- arbre prcdant le frre  supprimer
-- si l'exeption parente_vide est leve 
-- dans An_Frere, elle se propage
   a_pre:=An_Frere(a,n-1);
-- arbre suivant le frre  supprimer
-- si a_pre ou a_pre.frere vaut null, l'exception 
-- constraint_error est leve et elle se propage 
   a_s:=a_pre.frere.frere; 
-- cas o a_pre ou a_pre.frere diffrent de null    
   a_pre.frere.pere:=null;
   a_pre.frere.frere:=null;
   a_pre.frere:=a_s;
end if;
    
exception
   
 -- si a_pre ou a_pre.frere vaut null, on ne fait rien  
 when constraint_error => null;
                               
end An_Supprimer_Frere;

----------------------------------------------------------------
---- procdure An_Changer_Aux : procdure qui permet de parcourir
---- l'arbre et d'appliquer la fonction An_Changer  chaque lment
---- de l'arbre
---- paramtres: arbre n-aire a
---- post-conditions: si l'arbre est vide on ne fait rien
----------------------------------------------------------------

procedure An_Changer( a: in out arb_nr ) is

begin

if An_Vide(a) then
   null;
  else
   a.val:=fonction_main(a.val);
   An_Changer(a.premier_fils);
   An_Changer(a.frere);
end if;   

end An_Changer;


-----------------------------------------------------------
---- procdure Inserer_Premier_Fils: insre un premier_fils
---- dans un arbre a.
---- paramtres: arbre n-aire a
----             arbre n-aire  insrer a_ins
-----------------------------------------------------------  
  

procedure Inserer_Premier_Fils ( a: in out arb_nr ; a_ins: in out arb_nr) is

a_s: arb_nr;

begin

a_s:=a.premier_fils;
a.premier_fils:=a_ins;
a_ins.pere:=a;
a_ins.premier_fils:=a_s;
a_s.pere:=a_ins;

end Inserer_Premier_Fils;

end arbre_nr;




 









